package ru.dbotthepony.mc.otm.client

import com.mojang.blaze3d.platform.InputConstants
import com.mojang.blaze3d.systems.RenderSystem
import net.minecraft.client.KeyMapping
import net.minecraftforge.client.event.RegisterKeyMappingsEvent
import net.minecraftforge.client.event.RenderGuiEvent
import net.minecraftforge.client.settings.KeyConflictContext
import org.lwjgl.glfw.GLFW.GLFW_CURSOR_DISABLED
import org.lwjgl.glfw.GLFW.GLFW_CURSOR_NORMAL
import ru.dbotthepony.mc.otm.OverdriveThatMatters
import ru.dbotthepony.mc.otm.android.AndroidFeatureType
import ru.dbotthepony.mc.otm.android.AndroidSwitchableFeature
import ru.dbotthepony.mc.otm.capability.matteryPlayer
import ru.dbotthepony.mc.otm.client.render.drawArc
import ru.dbotthepony.mc.otm.network.MatteryPlayerNetworkChannel
import ru.dbotthepony.mc.otm.network.SwitchAndroidFeaturePacket
import java.util.stream.Collectors
import kotlin.math.PI
import kotlin.math.acos
import kotlin.math.cos
import kotlin.math.pow
import kotlin.math.sin

// TODO: forge's KeyMapping patch (ALL and MAP maps) is not thread safe... should we be worried?
object AndroidMenuKeyMapping : KeyMapping("key.otm.android_menu", KeyConflictContext.IN_GAME, InputConstants.Type.KEYSYM.getOrCreate(InputConstants.KEY_F), OverdriveThatMatters.MOD_ID) {
	private var grabbedInput = false
	private var selectedFeature: AndroidSwitchableFeature? = null

	override fun setDown(isDown: Boolean) {
		val old = this.isDown
		super.setDown(isDown)

		if (old != isDown) {
			if (
				minecraft.screen == null &&
				isDown &&
				minecraft.player?.matteryPlayer?.isAndroid == true &&
				minecraft.player?.matteryPlayer?.features?.anyMatch { it is AndroidSwitchableFeature && it.allowToSwitchByPlayer } == true
			) {
				grabbedInput = true
				minecraft.mouseHandler.releaseMouse()
			} else if (!isDown && grabbedInput) {
				grabbedInput = false

				if (minecraft.screen == null) {
					minecraft.mouseHandler.grabMouse()

					val selectedFeature = selectedFeature
					if (selectedFeature != null) {
						MatteryPlayerNetworkChannel.sendToServer(SwitchAndroidFeaturePacket(selectedFeature.type, !selectedFeature.isActive))
					}
				}
			}

			selectedFeature = null
		}
	}

	fun onRenderGuiEvent(event: RenderGuiEvent.Post) {
		if (!grabbedInput) {
			return
		}

		val features = minecraft.player?.matteryPlayer?.features?.filter { it is AndroidSwitchableFeature && it.allowToSwitchByPlayer }?.collect(Collectors.toList()) as MutableList<AndroidSwitchableFeature>? ?: return

		if (features.isEmpty()) {
			return
		}

		RenderSystem.setShaderColor(0f, 0f, 0f, 0.75f)

		val size = minecraft.window.guiScaledHeight.coerceAtMost(minecraft.window.guiScaledWidth).toFloat() * 0.35f

		drawArc(
			event.poseStack,
			minecraft.window.guiScaledWidth / 2f,
			minecraft.window.guiScaledHeight / 2f,
			size,
			size * 0.3f
		)

		RenderSystem.setShaderColor(85 / 255f, 197 / 255f, 255 / 255f, 0.5f)

		val degreePerSlice = PI * 2.0 / features.size

		var (mouseX, mouseY) = mousePos

		mouseX -= minecraft.window.width / 2.0
		mouseY -= minecraft.window.height / 2.0

		//font.drawAligned(event.poseStack, mouseX.toString(), TextAlign.CENTER_CENTER, 100f, 40f, RGBAColor.WHITE)
		//font.drawAligned(event.poseStack, mouseY.toString(), TextAlign.CENTER_CENTER, 100f, 60f, RGBAColor.WHITE)

		val length = (mouseX * mouseX + mouseY * mouseY).pow(0.5)

		if ((length / minecraft.window.guiScale) in (size * 0.3f) .. size) {
			var deg = acos(mouseX / length)

			// opengl
			// lower values are at bottom of screen
			if (mouseY > 0.0) {
				deg = -deg
			}

			if (deg < 0.0) {
				deg += PI * 2
			}

			//font.drawAligned(event.poseStack, deg.toString(), TextAlign.CENTER_CENTER, 100f, 20f, RGBAColor.WHITE)

			val index = (deg / degreePerSlice).toInt()
			selectedFeature = features[index]

			//font.drawAligned(event.poseStack, index.toString(), TextAlign.CENTER_CENTER, 100f, 80f, RGBAColor.WHITE)

			RenderSystem.setShaderColor(85 / 255f, 197 / 255f, 255 / 255f, 0.3f)

			drawArc(
				event.poseStack,
				minecraft.window.guiScaledWidth / 2f,
				minecraft.window.guiScaledHeight / 2f,
				size * 1.2f,
				size * 0.4f,

				startDegree = index * degreePerSlice,
				endDegree = (index + 1) * degreePerSlice
			)

			//font.drawAligned(event.poseStack, (index * degreePerSlice).toString(), TextAlign.CENTER_CENTER, 100f, 100f, RGBAColor.WHITE)
		} else {
			selectedFeature = null
		}

		RenderSystem.setShaderColor(1f, 1f, 1f, 1f)

		val iconSize = size * 0.25f

		event.poseStack.pushPose()
		event.poseStack.translate(minecraft.window.guiScaledWidth.toDouble() / 2f, minecraft.window.guiScaledHeight.toDouble() / 2f, 0.0)

		if (features.size == 1) {
			val feature = features.first()

			feature.renderIcon(event.poseStack, -iconSize / 2f, -size * 0.7f - iconSize / 2f, iconSize, iconSize)
		} else {
			for ((index, feature) in features.withIndex()) {
				val sin = sin((index + 0.5) * degreePerSlice).toFloat()
				val cos = cos((index + 0.5) * degreePerSlice).toFloat()

				feature.renderIcon(event.poseStack, -iconSize / 2f + size * 0.7f * cos, -size * 0.7f * sin - iconSize / 2f, iconSize, iconSize)
			}
		}

		event.poseStack.popPose()
	}

	fun register(event: RegisterKeyMappingsEvent) {
		event.register(this)
	}
}
