package ru.dbotthepony.mc.otm.data

import com.mojang.datafixers.util.Pair
import com.mojang.serialization.Codec
import com.mojang.serialization.DataResult
import com.mojang.serialization.DynamicOps
import com.mojang.serialization.codecs.ListCodec
import net.minecraft.resources.ResourceLocation
import net.minecraft.world.item.ItemStack
import net.minecraft.world.item.Items
import net.minecraftforge.registries.ForgeRegistries

object ItemStackCodec : Codec<ItemStack> {
	override fun <T : Any?> encode(input: ItemStack, ops: DynamicOps<T>, prefix: T): DataResult<T> {
		require(prefix == ops.empty()) { "Non-empty prefix: $prefix" }

		return ForgeRegistries.ITEMS.codec.encode(input.item, ops, ops.empty()).map {
			ops.createMap(mapOf(
				ops.createString("id") to it,
				ops.createString("count") to ops.createInt(input.count)
			))
		}
	}

	override fun <T : Any?> decode(ops: DynamicOps<T>, input: T): DataResult<Pair<ItemStack, T>> {
		return ops.getMap(input).flatMap {
			val item = it["id"]?.let { ForgeRegistries.ITEMS.codec.decode(ops, it) }?.result()?.orElse(null)?.first
			val count = it["count"]?.let(ops::getNumberValue)?.result()?.orElse(null)?.toInt() ?: return@flatMap DataResult.error("Invalid item count")

			if (item == null || item == Items.AIR) {
				return@flatMap DataResult.error("Unknown item type $item")
			}

			DataResult.success(ItemStack(item, count))
		}.map { Pair.of(it, ops.empty()) }
	}

	val LIST = ListCodec(this)
}
